/* finish_sum */

#ifdef HAVE_CL_DOUBLE
#pragma OPENCL EXTENSION cl_khr_fp64 : enable
typedef double2 real2;
typedef double4 real4;
#define ZERO 0.0
#else
typedef float2  real2;
typedef float4  real4;
#define ZERO 0.0f
#endif


__kernel void finish_sum(
        const unsigned int   isize,
        const unsigned int   isize_old,
        const unsigned int   ioffset,
        __global const real4* state_old,
        __local        real2* tile)
{       
         unsigned int giX  = get_global_id(0);
   const unsigned int tiX  = get_local_id(0);

   const unsigned int ngX  = get_global_size(0);
   const unsigned int ntX  = get_local_size(0);

   tile[tiX].s0 = ZERO;
   tile[tiX].s1 = -INFINITY;
   while (giX < isize) {
     tile[tiX].s0 += state_old[giX].s3;
     if (state_old[ioffset+giX].s3 > tile[tiX].s1) tile[tiX].s1 = state_old[ioffset+giX].s3;
     giX += ngX;
   }
 
   barrier(CLK_LOCAL_MEM_FENCE);
 
   for (int offset=ntX>>1; offset > 32; offset >>= 1){
     if (tiX < offset){
       tile[tiX].s0 += tile[tiX+offset].s0;
       if (tile[tiX+offset].s1 > tile[tiX].s1) tile[tiX].s1 = tile[tiX+offset].s1;
     }
     barrier(CLK_LOCAL_MEM_FENCE);
   }
   if (tiX < 32){
     tile[tiX].s0 += tile[tiX+32].s0;
     tile[tiX].s0 += tile[tiX+16].s0;
     tile[tiX].s0 += tile[tiX+8].s0;
     tile[tiX].s0 += tile[tiX+4].s0;
     tile[tiX].s0 += tile[tiX+2].s0;
     tile[tiX].s0 += tile[tiX+1].s0;
     if (tile[tiX+32].s1 > tile[tiX].s1) tile[tiX].s1 = tile[tiX+32].s1;
     if (tile[tiX+16].s1 > tile[tiX].s1) tile[tiX].s1 = tile[tiX+16].s1;
     if (tile[tiX+8].s1 > tile[tiX].s1) tile[tiX].s1 = tile[tiX+8].s1;
     if (tile[tiX+4].s1 > tile[tiX].s1) tile[tiX].s1 = tile[tiX+4].s1;
     if (tile[tiX+2].s1 > tile[tiX].s1) tile[tiX].s1 = tile[tiX+2].s1;
     if (tile[tiX+1].s1 > tile[tiX].s1) tile[tiX].s1 = tile[tiX+1].s1;
   }

   if (tiX == 0) {
     state_old[0].s3 = tile[tiX].s0;
     state_old[1].s3 = sqrt(tile[tiX].s1);
   }
}

