/*
 *  mesh.h
 *
 *  Copyright 2007-2010 Jon Robey and Bob Robey, brobey@lanl.gov 
 *  All rights reserved.
 *  Los Alamos National Laboratory and Supercomputing Challenge
 *  Permission granted for use in Supercomputing Challenge Projects
 */

#ifndef MESH_H_
#define MESH_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifndef EZCL_H_
#ifdef HAVE_CL_DOUBLE
#pragma OPENCL EXTENSION cl_khr_fp64 : enable
typedef double      real;
typedef cl_double   cl_real;
typedef cl_double4  cl_real4;
#else
typedef float       real;
typedef cl_float    cl_real;
typedef cl_float4   cl_real4;
#endif
#endif

#define SLOPED_WAVE 1
#define WAVE_BOX    2

void mesh_init(unsigned int problem, unsigned int imax, unsigned int jmax,
            unsigned int imax_padded, cl_real4** state);
double mesh_get_deltaX(void);
double mesh_get_deltaY(void);
double mesh_get_xmin(void);
double mesh_get_xmax(void);
double mesh_get_ymin(void);
double mesh_get_ymax(void);
double *mesh_get_x(void);
double *mesh_get_y(void);
double *mesh_get_dx(void);
double *mesh_get_dy(void);

void mesh_mem_zero_init(int jsize, int isize, cl_real4** state);
double mesh_sum_var(cl_real4** state);
void mesh_set_plot_var(cl_real4** state, double **plot_var);
double mesh_set_timestep(double sigma, cl_real4** state);
void mesh_output(char *title, cl_real4** state);

#endif /* MESH_H_ */
