/* wave_kern.cl */

#ifdef HAVE_CL_DOUBLE
#pragma OPENCL EXTENSION cl_khr_fp64 : enable
typedef double  real;
typedef double4 real4;
typedef double8 real8;
#define _REAL_CONSTANT
#define ZERO 0.0
#define HALF 0.5
#define ONE  1.0
#define GRAVITATIONAL_CONSTANT  9.8
#else
typedef float   real;
typedef float4  real4;
typedef float8  real8;
#define _REAL_CONSTANT f
#define ZERO 0.0f
#define HALF 0.5f
#define ONE  1.0f
#define GRAVITATIONAL_CONSTANT  9.8f
#endif

#define HXFLUX(i) ( Uval(i) )
#define UXFLUX(i) ( SQ(Uval(i))/Hval(i) + ghalf*SQ(Hval(i)) )
#define VXFLUX(i) ( Uval(i)*Vval(i)/Hval(i) )

#define HYFLUX(i) ( Vval(i) )
#define UYFLUX(i) ( Uval(i)*Vval(i)/Hval(i) )
#define VYFLUX(i) ( SQ(Vval(i))/Hval(i) + ghalf*SQ(Hval(i)) )

#define HNEWXFLUX(i)  ( UxPanel(i) )
#define HNEWYFLUX(i)  ( VyPanel(i) )
#define UNEWXFLUX(i)  ( SQ(UxPanel(i) )/HxPanel(i)  + ghalf*SQ(HxPanel(i) ) )
#define UNEWCFLUX(i)  ( VyPanel(i) *UyPanel(i) /HyPanel(i)  )
#define VNEWYFLUX(i)  ( SQ(VyPanel(i) )/HyPanel(i)  + ghalf*SQ(HyPanel(i) ) )
#define VNEWCFLUX(i)  ( UxPanel(i) *VxPanel(i) /HxPanel(i)  )

#define SQ(x)    ((x)*(x))
//define macro to find minimum of 3 values
#define MIN3(a,b,c) (min(min((a),(b)),(c)))

#define Hval(i)  (tile[i].s0)
#define Uval(i)  (tile[i].s1)
#define Vval(i)  (tile[i].s2)
#define Wval(i)  (tile[i].s3)

#define Hnew(i)  (state_new[i].s0)
#define Unew(i)  (state_new[i].s1)
#define Vnew(i)  (state_new[i].s2)

#define HxPanel(i) (XPanel[i][0])
#define UxPanel(i) (XPanel[i][1])
#define VxPanel(i) (XPanel[i][2])

#define HyPanel(i) (YPanel[i][0])
#define UyPanel(i) (YPanel[i][1])
#define VyPanel(i) (YPanel[i][2])

__kernel void wave_kern(
        const real deltaT,
        const real deltaX,
        const real deltaY,
        const unsigned int   isize,
        const unsigned int   jsize,
        __global const real4* state_old,
        __global       real4* state_new,
        __local        real4* tile)
{       
   const unsigned int giX  = get_global_id(0);
   const unsigned int tiX  = get_local_id(0);

   const unsigned int ngX  = get_global_size(0);
   const unsigned int ntX  = get_local_size(0);

   const unsigned int giY  = get_global_id(1);
   const unsigned int tiY  = get_local_id(1);

   const unsigned int ngY  = get_global_size(1);
   const unsigned int ntY = get_local_size(1);
   
   const unsigned int group_id = get_group_id(0);

   const unsigned int stridetY = ntX+4;
   
   const unsigned int ic  = (tiX+2)+(tiY+2)*stridetY; // Current tile index
   const unsigned int il  = ic-1;        // Index one cell  to the left
   const unsigned int ir  = ic+1;        // Index one cell  to the right
   const unsigned int ib  = ic-stridetY; // Index one cell  to the bottom
   const unsigned int it  = ic+stridetY; // Index one cell  to the top
   const unsigned int ill = il-1;        // Index two cells to the left
   const unsigned int irr = ir+1;        // Index two cells to the right
   const unsigned int ibb = ib-stridetY; // Index two cells to the bottom
   const unsigned int itt = it+stridetY; // Index two cells to the top
   
   //const unsigned int tid = tiX+tiY*ntX;

   const unsigned int strideY = ngX;
   const unsigned int gindex = giX+giY*strideY;                               

   tile[ic]=state_old[gindex];
   
// Reflective boundary conditions
   if (giX == 0)     {
     Hval(il) =  Hval(ic);
     Uval(il) =(-Uval(ic));
     Vval(il) =  Vval(ic);
     Hval(ill)=  Hval(ic);
     Uval(ill)=(-Uval(ic));
     Vval(ill)=  Vval(ic);
   }
   else if (tiX == 0)     {
     tile[il] =state_old[gindex-1];
     tile[ill]=state_old[gindex-1];
   }
   if (giX == isize-1) {
     Hval(ir) =  Hval(ic);
     Uval(ir) =(-Uval(ic));
     Vval(ir) =  Vval(ic);
     Hval(irr)=  Hval(ic);
     Uval(irr)=(-Uval(ic));
     Vval(irr)=  Vval(ic);
   }
   else if (tiX == ntX-1) {
     tile[ir] =state_old[gindex+1];
     tile[irr]=state_old[gindex+1];
   }
   if (giY == 0)     {
     Hval(ib) =  Hval(ic);
     Uval(ib) =  Uval(ic);
     Vval(ib) =(-Vval(ic));
     Hval(ibb)=  Hval(ic);
     Uval(ibb)=  Uval(ic);
     Vval(ibb)=(-Vval(ic));
   }
   else if (tiY == 0)     {
     tile[ib] =state_old[gindex-strideY];
     tile[ibb]=state_old[gindex-strideY];
   }
   if (giY == jsize-1) {
     Hval(it) =  Hval(ic);
     Uval(it) =  Uval(ic);
     Vval(it) =(-Vval(ic));
     Hval(itt)=  Hval(ic);
     Uval(itt)=  Uval(ic);
     Vval(itt)=(-Vval(ic));
   }
   else if (tiY == ntY-1) {
     tile[it] =state_old[gindex+strideY];
     tile[itt]=state_old[gindex+strideY];
   }
   
   barrier(CLK_LOCAL_MEM_FENCE);

   const real Cx     = deltaT/deltaX;
   const real Cxhalf = HALF*Cx;
   const real Cy     = deltaT/deltaY;
   const real Cyhalf = HALF*Cy;
   const real g      = GRAVITATIONAL_CONSTANT;
   const real ghalf  = HALF*g;
   real XPanel[2][3];
   real YPanel[2][3];
      
   HxPanel(0) = HALF*(Hval(ic) + Hval(il)) - Cxhalf * ( HXFLUX(ic) - HXFLUX(il) );
   UxPanel(0) = HALF*(Uval(ic) + Uval(il)) - Cxhalf * ( UXFLUX(ic) - UXFLUX(il) );
   VxPanel(0) = HALF*(Vval(ic) + Vval(il)) - Cxhalf * ( VXFLUX(ic) - VXFLUX(il) );

   HxPanel(1) = HALF*(Hval(ir) + Hval(ic)) - Cxhalf * ( HXFLUX(ir) - HXFLUX(ic) );
   UxPanel(1) = HALF*(Uval(ir) + Uval(ic)) - Cxhalf * ( UXFLUX(ir) - UXFLUX(ic) );
   VxPanel(1) = HALF*(Vval(ir) + Vval(ic)) - Cxhalf * ( VXFLUX(ir) - VXFLUX(ic) );
                                                              
   HyPanel(0) = HALF*(Hval(ic) + Hval(ib)) - Cyhalf * ( HYFLUX(ic) - HYFLUX(ib) );
   UyPanel(0) = HALF*(Uval(ic) + Uval(ib)) - Cyhalf * ( UYFLUX(ic) - UYFLUX(ib) );
   VyPanel(0) = HALF*(Vval(ic) + Vval(ib)) - Cyhalf * ( VYFLUX(ic) - VYFLUX(ib) );
   
   HyPanel(1) = HALF*(Hval(it) + Hval(ic)) - Cyhalf * ( HYFLUX(it) - HYFLUX(ic) );
   UyPanel(1) = HALF*(Uval(it) + Uval(ic)) - Cyhalf * ( UYFLUX(it) - UYFLUX(ic) );
   VyPanel(1) = HALF*(Vval(it) + Vval(ic)) - Cyhalf * ( VYFLUX(it) - VYFLUX(ic) );

   real duminus1, duminus2;
   real duplus1,  duplus2;
   real duhalf1,  duhalf2;
   real rdenom, rnumplus, rnumminus, rminus, rplus;
   real nu, wminusx, wplusx, wminusy, wplusy;

   duminus1 = Hval(il)-Hval(ill);
   duminus2 = Uval(il)-Uval(ill);
   duplus1  = Hval(ir)-Hval(ic);
   duplus2  = Uval(ir)-Uval(ic);
   duhalf1  = Hval(ic)-Hval(il);
   duhalf2  = Uval(ic)-Uval(il);
   rdenom = max(SQ(duhalf1) + SQ(duhalf2),ONE-30);
   rnumplus  = duplus1 *duhalf1 + duplus2 *duhalf2;
   rnumminus = duminus1*duhalf1 + duminus2*duhalf2;
   rplus  = rnumplus /rdenom;
   rminus = rnumminus/rdenom;
   nu=(fabs(UxPanel(0))+sqrt(g*HxPanel(0)))*Cx;
   wminusx = HALF*nu*(ONE-nu)*(ONE-max(MIN3(ONE, rminus, rplus), ZERO));

   duminus1 = Hval(ic)-Hval(il);
   duminus2 = Uval(ic)-Uval(il);
   duplus1  = Hval(irr)-Hval(ir);
   duplus2  = Uval(irr)-Uval(ir);
   duhalf1  = Hval(ir)-Hval(ic);
   duhalf2  = Uval(ir)-Uval(ic);
   rdenom = max(SQ(duhalf1) + SQ(duhalf2),ONE-30);
   rnumplus  = duplus1 *duhalf1 + duplus2 *duhalf2;
   rnumminus = duminus1*duhalf1 + duminus2*duhalf2;
   rplus  = rnumplus /rdenom;
   rminus = rnumminus/rdenom;
   nu=(fabs(UxPanel(1))+sqrt(g*HxPanel(1)))*Cx;
   wplusx = HALF*nu*(ONE-nu)*(ONE-max(MIN3(ONE, rminus, rplus), ZERO));

   duminus1 = Hval(ib)-Hval(ibb);
   duminus2 = Vval(ib)-Vval(ibb);
   duplus1  = Hval(it)-Hval(ic);
   duplus2  = Vval(it)-Vval(ic);
   duhalf1  = Hval(ic)-Hval(ib);
   duhalf2  = Vval(ic)-Vval(ib);
   rdenom = max(SQ(duhalf1) + SQ(duhalf2),ONE-30);
   rnumplus  = duplus1 *duhalf1 + duplus2 *duhalf2;
   rnumminus = duminus1*duhalf1 + duminus2*duhalf2;
   rplus  = rnumplus /rdenom;
   rminus = rnumminus/rdenom;
   nu=(fabs(VyPanel(0))+sqrt(g*HyPanel(0)))*Cy;
   wminusy = HALF*nu*(ONE-nu)*(ONE-max(MIN3(ONE, rminus, rplus), ZERO));

   duminus1 = Hval(ic)-Hval(ib);
   duminus2 = Vval(ic)-Vval(ib);
   duplus1  = Hval(itt)-Hval(it);
   duplus2  = Vval(itt)-Vval(it);
   duhalf1  = Hval(it)-Hval(ic);
   duhalf2  = Vval(it)-Vval(ic);
   rdenom = max(SQ(duhalf1) + SQ(duhalf2),ONE-30);
   rnumplus  = duplus1 *duhalf1 + duplus2 *duhalf2;
   rnumminus = duminus1*duhalf1 + duminus2*duhalf2;
   rplus  = rnumplus /rdenom;
   rminus = rnumminus/rdenom;
   nu=(fabs(VyPanel(1))+sqrt(g*HyPanel(1)))*Cy;
   wplusy = HALF*nu*(ONE-nu)*(ONE-max(MIN3(ONE, rminus, rplus), ZERO));

// The if block is to avoid returning values for the padded memory
   if (giX < isize && giY < jsize) {    
     real4 out;
     out = tile[ic];                                                  
     out.s0 +=  - Cx * ( HNEWXFLUX(1) - HNEWXFLUX(0) ) -wminusx*(Hval(ic)-Hval(il))+wplusx*(Hval(ir)-Hval(ic));
     out.s0 +=  - Cy * ( HNEWYFLUX(1) - HNEWYFLUX(0) ) -wminusy*(Hval(ic)-Hval(ib))+wplusy*(Hval(it)-Hval(ic));
     out.s1 +=  - Cx * ( UNEWXFLUX(1) - UNEWXFLUX(0) ) -wminusx*(Uval(ic)-Uval(il))+wplusx*(Uval(ir)-Uval(ic));
     out.s1 +=  - Cy * ( UNEWCFLUX(1) - UNEWCFLUX(0) ) -wminusy*(Uval(ic)-Uval(ib))+wplusy*(Uval(it)-Uval(ic));
     out.s2 +=  - Cx * ( VNEWCFLUX(1) - VNEWCFLUX(0) ) -wminusx*(Vval(ic)-Vval(il))+wplusx*(Vval(ir)-Vval(ic));
     out.s2 +=  - Cy * ( VNEWYFLUX(1) - VNEWYFLUX(0) ) -wminusy*(Vval(ic)-Vval(ib))+wplusy*(Vval(it)-Vval(ic));
     out.s3 = sqrt(GRAVITATIONAL_CONSTANT*Hval(ic));
     out.s3 = SQ( (out.s3+fabs(out.s1))/deltaX ) + SQ( (out.s3+fabs(out.s2))/deltaY );
     out.s3 = sqrt( SQ( (out.s3+fabs(out.s1))/deltaX ) + SQ( (out.s3+fabs(out.s2))/deltaY ) );
     state_new[gindex].s0123 = out.s0123;
     barrier(CLK_LOCAL_MEM_FENCE);
     tile[ic]=out;
   } 
   barrier(CLK_LOCAL_MEM_FENCE);
   real sum = ZERO;
   int isum;
   real localmax = -INFINITY;
   for (int offset=ntY>>1; offset > 0; offset >>= 1){
     if (tiY < offset){
       isum = (tiX+2)+(tiY+offset+2)*stridetY;
       Hval(ic) += Hval(isum);
       if (Wval(isum) > Wval(ic)) Wval(ic) = Wval(isum);
     }
     barrier(CLK_LOCAL_MEM_FENCE);
   }
   for (int offset=ntX>>1; offset > 0; offset >>= 1){
     if (tiY == 0 && tiX < offset){
       isum = (tiX+offset+2)+2*stridetY;
       Hval(ic) += Hval(isum);
       if (Wval(isum) > Wval(ic)) Wval(ic) = Wval(isum);
     }
     barrier(CLK_LOCAL_MEM_FENCE);
   }
   if (tiY == 0 && tiX == 0) {
     state_new[get_group_id(0)+get_group_id(1)*ngX/ntX].s3 = Hval(ic);
     state_new[(ngX/ntX)*(ngY/ntY) + get_group_id(0)+get_group_id(1)*ngX/ntX].s3 = Wval(ic);
   }
}



